header {* Cardinal Arithmetic  *}

(* author: Dmitriy Traytel *)

theory Cardinal_Arithmetic
imports Cardinal_Order_Relation
begin

text {*
  The following collection of lemmas should be seen as an user interface to the HOL Theory
  of cardinals. It is not expected to be complete in any sense, since it's
  development was driven by demand arising from the development of the (co)datatype package.
*}

(*library candidate*)
lemma dir_image: "\<lbrakk>\<And>x y. (f x = f y) = (x = y); Card_order r\<rbrakk> \<Longrightarrow> r =o dir_image r f"
by (rule dir_image_ordIso) (auto simp add: inj_on_def)

(*should supersede a weaker lemma from the library*)
lemma dir_image_Field: "Field (dir_image r f) = f ` Field r"
unfolding dir_image_def Field_def Range_def Domain_def by auto

lemma card_order_dir_image: 
  assumes bij: "bij f" and co: "card_order r"
  shows "card_order (dir_image r f)"
proof -
  from assms have "Field (dir_image r f) = UNIV"
    using card_order_on_Card_order[of UNIV r] unfolding bij_def dir_image_Field by auto
  moreover from bij have "\<And>x y. (f x = f y) = (x = y)" unfolding bij_def inj_on_def by auto
  with co have "Card_order (dir_image r f)" 
    using card_order_on_Card_order[of UNIV r] Card_order_ordIso2[OF _ dir_image] by blast
  ultimately show ?thesis by auto
qed

(*library candidate*)
lemma ordIso_refl: "Card_order r \<Longrightarrow> r =o r"
by (rule card_order_on_ordIso) assumption

(*library candidate*)
lemma ordLeq_refl: "Card_order r \<Longrightarrow> r \<le>o r"
by (rule ordIso_imp_ordLeq, rule card_order_on_ordIso) assumption

(*library candidate*)
lemma card_of_refl_ordLeq: "|A| \<le>o |A|"
by simp

(*library candidate*)
lemma card_of_ordIso_subst: "A = B \<Longrightarrow> |A| =o |B|"
by (simp only: ordIso_refl card_of_Card_order)

(*library candidate*)
lemma card_of_Ball: "|{x \<in> A. P x}| \<le>o |A|" (is "|?L| \<le>o |?R|")
proof -
  have "inj_on id ?L" unfolding inj_on_def by simp
  moreover have "id ` ?L \<subseteq> ?R" by auto
  ultimately show ?thesis unfolding card_of_ordLeq[symmetric] by blast
qed

(*library candidate*)
lemma ordIso_bij: "|A| =o |B| \<Longrightarrow> \<exists>f. bij_betw f A B"
unfolding ordIso_def iso_def_raw by auto

(*library candidate*)
lemma card_of_Times_Plus_distrib:
  "|A <*> (B <+> C)| =o |A <*> B <+> A <*> C|" (is "|?RHS| =o |?LHS|")
proof -
  let ?f = "\<lambda>(a, bc). case bc of Inl b \<Rightarrow> Inl (a, b) | Inr c \<Rightarrow> Inr (a, c)"
  have "bij_betw ?f ?RHS ?LHS" unfolding bij_betw_def inj_on_def by force
  thus ?thesis using card_of_ordIso by blast
qed

subsection {* Zero *}

definition czero where
  "czero = card_of {}"

lemma empty_czero[simp]:
  "|{}| = czero"
unfolding czero_def by simp

lemma czero_ordIso[simp]:
  "czero =o czero"
using card_of_empty_ordIso by simp

lemma empty_czero_ordIso:
  "|{}| =o czero"
using card_of_empty_ordIso by simp

lemma czero_ordLeq[simp]:
  "Card_order r \<Longrightarrow> czero \<le>o r"
unfolding czero_def using Card_order_empty .

lemma card_of_czero_iff_empty:
  "|A| = czero \<longleftrightarrow> A = {}"
proof
  assume "|A| = czero"
  hence "Field |A| = Field czero" by auto
  thus "A = {}" unfolding czero_def by (simp only: Field_card_of)
qed simp

lemma card_of_ordIso_czero_iff_empty:
  "|A| =o (czero :: 'a rel) \<longleftrightarrow> A = ({} :: 'a set)"
unfolding czero_def by (rule iffI[OF card_of_empty2]) auto

(* A "not czero" Cardinal predicate *)
abbreviation Cnotzero where
  "Cnotzero (r :: 'a rel) \<equiv> \<not>(r =o (czero :: 'a rel)) \<and> Card_order r"

(*helper*)
lemma Cnotzero_imp_not_empty:
  assumes "Cnotzero r"
  shows "Field r \<noteq> {}"
proof (rule ccontr)
  assume "\<not> (Field r \<noteq> {})"
  hence "|Field r| =o czero" by simp
  with assms(1) have "r =o czero" by (blast intro: card_of_unique ordIso_transitive)
  with assms show False by blast
qed

lemma czeroI:
  "\<lbrakk>Card_order r; Field r = {}\<rbrakk> \<Longrightarrow> r =o czero"
using Cnotzero_imp_not_empty ordIso_transitive[OF _ czero_ordIso] by blast

lemma czeroE:
  "r =o czero \<Longrightarrow> Field r = {}"
unfolding czero_def
by (drule card_of_cong) (simp only: Field_card_of card_of_empty2)

lemma Cnotzero_mono:
  "\<lbrakk>Cnotzero r; Card_order q; r \<le>o q\<rbrakk> \<Longrightarrow> Cnotzero q"
apply (rule ccontr)
apply auto
apply (drule czeroE)
apply (erule notE)
apply (erule czeroI)
apply (drule card_of_mono2)
apply (simp only: card_of_empty3)
done

subsection {* Infinite cardinals *}

definition cinfinite where
  "cinfinite r = infinite (Field r)"

abbreviation Cinfinite where
  "Cinfinite r \<equiv> cinfinite r \<and> Card_order r"

lemma natLeq_ordLeq_cinfinite:
  assumes inf: "Cinfinite r"
  shows "natLeq \<le>o r"
proof -
  from inf have "natLeq \<le>o |Field r|" by (simp add: cinfinite_def infinite_iff_natLeq_ordLeq)
  also from inf have "|Field r| =o r" by simp
  finally show ?thesis .
qed

lemma cinfinite_not_czero[simp]:
  "cinfinite r \<Longrightarrow> \<not> (r =o (czero :: 'a rel))"
proof
  assume "cinfinite r" "r =o (czero :: 'a rel)"
  then obtain f where "bij_betw f (Field r) (Field (czero :: 'a rel))"
    unfolding ordIso_def iso_def_raw by auto
  hence "bij_betw f (Field r) {}" unfolding czero_def by (simp only: Field_card_of)
  hence "Field r = {}" using Fun.bij_betw_empty2 by blast
  with `cinfinite r` show False unfolding cinfinite_def by simp
qed

lemma Cinfinite_Cnotzero[simp]:
  "Cinfinite r \<Longrightarrow> Cnotzero r"
by simp

lemma Cinfinite_cong:
  assumes "r1 =o r2" "Cinfinite r1"
  shows "Cinfinite r2"
proof (unfold cinfinite_def, rule ccontr)
  assume "\<not> (infinite (Field r2) \<and> Card_order r2)"
  hence "finite (Field r2) \<or> \<not> Card_order r2" by simp
  with assms have "finite (Field r2)" using Card_order_ordIso2 by blast
  hence "finite (Field r1)"
    using card_of_ordIso_finite[OF card_of_cong[OF assms(1)]] by simp
  with assms(2) show False unfolding cinfinite_def by simp
qed

lemma cinfinite_mono:
  assumes "r1 \<le>o r2" "cinfinite r1"
  shows "cinfinite r2"
proof (unfold cinfinite_def, rule ccontr)
  assume "\<not> infinite (Field r2)"
  hence "finite (Field r2)" by simp
  hence "finite (Field r1)"
    using card_of_ordLeq_finite[OF card_of_mono2[OF assms(1)]] by simp
  with assms(2) show False unfolding cinfinite_def by simp
qed

lemma card_of_Cinfinite_mono:
  "\<lbrakk>r \<le>o |A|; Cinfinite r\<rbrakk> \<Longrightarrow> Cinfinite |A|"
using cinfinite_mono card_of_Card_order by (rule conjI) auto

lemma Cinfinite_mono:
  "\<lbrakk>r1 \<le>o r2; Cinfinite r1; Card_order r2\<rbrakk> \<Longrightarrow> Cinfinite r2"
using cinfinite_mono by (rule conjI) auto


subsection {* Binary sum *}

definition csum (infixr "+c" 65) where
  "r1 +c r2 \<equiv> |Field r1 <+> Field r2|"

lemma Card_order_csum[simp]:
  "Card_order ( r1 +c r2 )"
unfolding csum_def by simp

lemma csum_not_czero[simp]:
  "A \<noteq> {} \<or> B \<noteq> {} \<Longrightarrow> \<not> ( |A| +c |B| =o czero)"
unfolding czero_def csum_def
using Field_card_of Plus_eq_empty_conv card_of_empty2 by auto

lemma csum_Cnotzero:
  "A \<noteq> {} \<or> B \<noteq> {} \<Longrightarrow> Cnotzero ( |A| +c |B| )"
unfolding czero_def csum_def
using Field_card_of Plus_eq_empty_conv card_of_empty2 by auto

lemma csum_not_czero1[simp]:
  assumes "Cnotzero r1"
  shows "\<not> ( r1 +c r2 =o czero)"
proof -
  from assms(1) have *: "|Field r1| =o r1" by simp
  moreover
  from assms have "Field r1 \<noteq> {}" using Cnotzero_imp_not_empty[of r1] by simp
  hence "\<not> ( |Field r1| +c |Field r2| =o czero)" by simp
  ultimately show ?thesis by (simp add: csum_def)
qed

lemma csum_Cnotzero1:
  "Cnotzero r1 \<Longrightarrow> Cnotzero (r1 +c r2)"
by simp

lemma csum_not_czero2[simp]:
  assumes "Cnotzero r2"
  shows "\<not> ( r1 +c r2 =o czero)"
proof -
  from assms(1) have *: "|Field r2| =o r2" by auto
  moreover
  from assms have "Field r2 \<noteq> {}" using Cnotzero_imp_not_empty[of r2] by auto
  hence "\<not> ( |Field r1| +c |Field r2| =o czero)" by simp
  ultimately show ?thesis by (simp add: csum_def)
qed

lemma csum_Cnotzero2:
  "Cnotzero r2 \<Longrightarrow> Cnotzero (r1 +c r2)"
by simp

lemma csum_not_czero'[simp]:
  assumes "Cnotzero r1 \<or> Cnotzero r2"
  shows "\<not> ( r1 +c r2 =o czero)"
proof -
  from assms have "Field r1 \<noteq> {} \<or> Field r2 \<noteq> {}"
    using Cnotzero_imp_not_empty[of r1] Cnotzero_imp_not_empty[of r2] by blast
  hence "\<not> ( |Field r1| +c |Field r2| =o czero)" by (rule csum_not_czero)
  thus ?thesis by (simp add: csum_def)
qed

lemma csum_Cnotzero':
  "Cnotzero r1 \<or> Cnotzero r2 \<Longrightarrow> Cnotzero (r1 +c r2)"
by simp

lemma card_order_csum[simp]:
  assumes "card_order r1" "card_order r2"
  shows "card_order (r1 +c r2)"
proof -
  have "Field r1 = UNIV" "Field r2 = UNIV" using assms card_order_on_Card_order by auto
  thus ?thesis unfolding csum_def by auto
qed

lemma cinfinite_csum[simp]:
  "cinfinite r1 \<or> cinfinite r2 \<Longrightarrow> cinfinite (r1 +c r2)"
unfolding cinfinite_def csum_def by auto

lemma cinfinite_csum1[simp]:
  "cinfinite r1 \<Longrightarrow> cinfinite (r1 +c r2)"
unfolding cinfinite_def csum_def by auto

lemma cinfinite_csum2[simp]:
  "cinfinite r2 \<Longrightarrow> cinfinite (r1 +c r2)"
unfolding cinfinite_def csum_def by auto

lemma Cinfinite_csum[simp]:
  "Cinfinite r1 \<or> Cinfinite r2 \<Longrightarrow> Cinfinite (r1 +c r2)"
unfolding cinfinite_def csum_def by auto

lemma Cinfinite_csum1[simp]:
  "Cinfinite r1 \<Longrightarrow> Cinfinite (r1 +c r2)"
unfolding cinfinite_def csum_def by auto

lemma Cinfinite_csum2[simp]:
  "Cinfinite r2 \<Longrightarrow> Cinfinite (r1 +c r2)"
unfolding cinfinite_def csum_def by auto

lemma csum_cong[simp]:
  assumes "p1 =o r1" and "p2 =o r2"
  shows "p1 +c p2 =o r1 +c r2"
unfolding csum_def by (simp only: assms ordIso_Plus_cong)

lemma csum_cong1[simp]:
  assumes "p1 =o r1"
  shows "p1 +c q =o r1 +c q"
unfolding csum_def by (simp only: assms ordIso_Plus_cong1)

lemma csum_cong2[simp]:
  assumes "p2 =o r2"
  shows "q +c p2 =o q +c r2"
unfolding csum_def by (simp only: assms ordIso_Plus_cong2)

lemma csum_mono[simp]:
  assumes "p1 \<le>o r1" and "p2 \<le>o r2"
  shows "p1 +c p2 \<le>o r1 +c r2"
unfolding csum_def by (simp only: assms ordLeq_Plus_mono)

lemma csum_mono1[simp]:
  assumes "p1 \<le>o r1"
  shows "p1 +c q \<le>o r1 +c q"
unfolding csum_def by (simp only: assms ordLeq_Plus_mono1)

lemma csum_mono2[simp]:
  assumes "p2 \<le>o r2"
  shows "q +c p2 \<le>o q +c r2"
unfolding csum_def by (simp only: assms ordLeq_Plus_mono2)

lemma ordLeq_csum1:
  assumes "Card_order p1"
  shows "p1 \<le>o p1 +c p2"
unfolding csum_def by (simp only: assms Card_order_Plus1)

lemma ordLeq_csum2:
  assumes "Card_order p2"
  shows "p2 \<le>o p1 +c p2"
unfolding csum_def by (simp only: assms Card_order_Plus2)

lemma csum_com:
  "p1 +c p2 =o p2 +c p1"
unfolding csum_def by (simp only: card_of_Plus_commute)

lemma csum_assoc:
  "(p1 +c p2) +c p3 =o p1 +c p2 +c p3"
unfolding csum_def by (simp only: Field_card_of card_of_Plus_assoc)

lemma Plus_csum[simp]:
  "|A <+> B| =o |A| +c |B|"
unfolding csum_def by (simp only: Field_card_of card_of_refl)

lemma czero_csum[simp]:
  assumes "Card_order r"
  shows "czero +c r =o r"
proof -
  have "|{} <+> Field r| =o r" by (simp only: Card_order_Plus_empty2 assms ordIso_symmetric)
  thus ?thesis unfolding csum_def czero_def by (simp only: Field_card_of)
qed

lemma csum_czero[simp]:
  assumes "Card_order r"
  shows "r +c czero =o r"
proof -
  have "|Field r <+> {}| =o r" by (simp only: Card_order_Plus_empty1 assms ordIso_symmetric)
  thus ?thesis unfolding csum_def czero_def by (simp only: Field_card_of)
qed

lemma Un_csum[simp]:
  "|A \<union> B| \<le>o |A| +c |B|"
using ordLeq_ordIso_trans[OF card_of_Un_Plus_ordLeq Plus_csum] by blast

lemmas Un_csum3[simp] =
  ordLeq_transitive[OF Un_csum csum_mono1[OF Un_csum]]
  ordLeq_transitive[OF Un_csum csum_mono2[OF Un_csum]]


subsection {* One *}

definition cone where
  "cone = card_of {()}"

lemma card_order_cone[simp]:
  "card_order cone"
unfolding cone_def using UNIV_unit by simp

lemma Card_order_cone[simp]:
  "Card_order cone"
unfolding cone_def by simp

lemma single_cone:
  "|{x}| =o cone"
proof -
  let ?f = "\<lambda>x. ()"
  have "bij_betw ?f {x} {()}" unfolding bij_betw_def by auto
  thus ?thesis unfolding cone_def using card_of_ordIso by blast
qed

lemma czero_not_cone:
  "\<not> (czero =o cone)"
using card_of_empty3[of "{()}"] ordIso_iff_ordLeq by (auto simp: cone_def)

lemma cone_not_czero:
  "\<not> (cone =o czero)"
using card_of_empty3[of "{()}"] ordIso_iff_ordLeq by (auto simp: cone_def)

lemma cone_Cnotzero:
  "Cnotzero cone"
by (simp add: cone_not_czero)

lemma cone_ordLeq_Cnotzero:
  assumes r: "Cnotzero r" (is "\<not> (_ =o ?zero) \<and> _")
  shows "cone \<le>o r"
proof -
  from r have "Field r \<noteq> {}" using czeroI by blast
  then obtain x where "x \<in> Field r" by blast
  hence "cone \<le>o |Field r|"
    unfolding cone_def inj_on_def card_of_ordLeq[symmetric]
    using exI[of _ "\<lambda>_. x"] by auto
  with r show ?thesis by (simp add: ordLeq_ordIso_trans)
qed


subsection{* Two *}

definition ctwo where
  "ctwo = |UNIV :: bool set|"

lemma Card_order_ctwo[simp]:
  "Card_order ctwo"
unfolding ctwo_def by simp

lemma ordLeq_ctwo:
  "|{a,b}| \<le>o ctwo"
proof -
  have ab: "{a,b} \<noteq> {}" by simp
  show ?thesis unfolding ctwo_def card_of_ordLeq2[OF ab, symmetric]
    by (rule exI[of _ "\<lambda> c. if c then a else b"]) auto
qed

lemma ordIso_ctwo:
  "a \<noteq> b \<Longrightarrow> |{a,b}| =o ctwo"
unfolding ctwo_def card_of_ordIso[symmetric] bij_betw_def
by (rule exI[of _ "\<lambda>c. c = a"]) auto

lemma cone_ordLeq_ctwo:
  "cone \<le>o ctwo"
unfolding cone_def ctwo_def card_of_ordLeq[symmetric] by auto

lemma cone_ordLess_ctwo:
  "cone <o ctwo"
proof-
  { assume "ctwo \<le>o cone"
    hence "|UNIV::bool set| \<le>o |UNIV::unit set|"
      unfolding ctwo_def cone_def by (auto intro: card_of_UNIV ordLeq_transitive)
    then obtain f::"bool \<Rightarrow> unit" where "inj f"
      unfolding card_of_ordLeq[symmetric] by auto
    hence "f True \<noteq> f False" unfolding inj_on_def by auto
    hence False by auto
  }
  thus ?thesis using cone_ordLeq_ctwo
    using ordIso_iff_ordLeq ordLeq_iff_ordLess_or_ordIso by auto
qed

lemma czero_not_ctwo:
  "\<not> (czero =o ctwo)"
using card_of_empty3[of "UNIV :: bool set"] ordIso_iff_ordLeq by (auto simp: ctwo_def)

lemma ctwo_not_czero:
  "\<not> (ctwo =o czero)"
using card_of_empty3[of "UNIV :: bool set"] ordIso_iff_ordLeq by (auto simp: ctwo_def)

lemma ctwo_Cnotzero:
  "Cnotzero ctwo"
by (simp add: ctwo_not_czero)


subsection {* Family sum *}

definition Csum where
  "Csum r rs \<equiv> |SIGMA i : Field r. Field (rs i)|"

(* Similar setup to the one for SIGMA from theory Big_Operators: *)
syntax "_Csum" ::
  "pttrn => ('a * 'a) set => 'b * 'b set => (('a * 'b) * ('a * 'b)) set"
  ("(3CSUM _:_. _)" [0, 51, 10] 10)

translations
  "CSUM i:r. rs" == "CONST Csum r (%i. rs)"

lemma card_of_UNION_csum[simp]:
  "|\<Union>i\<in>I. A i| \<le>o (CSUM i : |I|. |A i| )"
unfolding Csum_def by (simp add: card_of_UNION_Sigma)

lemma SIGMA_CSUM[simp]:
  "|SIGMA i : I. As i| = (CSUM i : |I|. |As i| )"
unfolding Csum_def by simp
(* NB: Always, under the cardinal operator,
operations on sets are reduced automatically to operations on cardinals.
This should make cardinal reasoning more direct and natural.  *)


subsection {* Product *}

definition cprod (infixr "*c" 80) where
  "r1 *c r2 = |Field r1 <*> Field r2|"

lemma Times_cprod: "|A \<times> B| =o |A| *c |B|"
unfolding cprod_def by (simp only: Field_card_of card_of_refl)

lemma card_order_cprod[simp]:
  assumes "card_order r1" "card_order r2"
  shows "card_order (r1 *c r2)"
proof -
  have "Field r1 = UNIV" "Field r2 = UNIV" using assms card_order_on_Card_order by auto
  thus ?thesis unfolding cprod_def by auto
qed

lemma Card_order_cprod[simp]:
  "Card_order (r1 *c r2)"
unfolding cprod_def by (simp only: Field_card_of card_of_card_order_on)

lemma cprod_cong[simp]:
  assumes "p1 =o r1" and "p2 =o r2"
  shows "p1 *c p2 =o r1 *c r2"
unfolding cprod_def by (simp only: assms ordIso_Times_cong)

lemma cprod_cong1[simp]:
  assumes "p1 =o r1"
  shows "p1 *c q =o r1 *c q"
unfolding cprod_def by (simp only: assms ordIso_Times_cong1)

lemma cprod_cong2[simp]:
  assumes "p2 =o r2"
  shows "q *c p2 =o q *c r2"
unfolding cprod_def by (simp only: assms ordIso_Times_cong2)

lemma cprod_mono[simp]:
  assumes "p1 \<le>o r1" and "p2 \<le>o r2"
  shows "p1 *c p2 \<le>o r1 *c r2"
unfolding cprod_def by (simp only: assms ordLeq_Times_mono)

lemma cprod_mono1[simp]:
  assumes "p1 \<le>o r1"
  shows "p1 *c q \<le>o r1 *c q"
unfolding cprod_def by (simp only: assms ordLeq_Times_mono1)

lemma cprod_mono2[simp]:
  assumes "p2 \<le>o r2"
  shows "q *c p2 \<le>o q *c r2"
unfolding cprod_def by (simp only: assms ordLeq_Times_mono2)

lemma ordLeq_cprod1[simp]:
  assumes "Card_order p1" "Cnotzero p2"
  shows "p1 \<le>o p1 *c p2"
proof -
  from assms(2) have *: "|Field p2| =o p2" by simp
  { assume "|Field p2| = czero"
    with * have "czero =o p2" using czero_ordIso ordIso_transitive by fastforce
    hence "p2 =o czero" using ordIso_symmetric by blast
    with assms(2) have False by blast
  }
  hence "|Field p2| \<noteq> czero" by blast
  hence "Field p2 \<noteq> {}" by (simp add: card_of_czero_iff_empty)
  with assms(1) show ?thesis unfolding cprod_def by (simp add: Card_order_Times1 del: SIGMA_CSUM)
qed

lemma ordLeq_cprod1'[simp]:
  assumes "Card_order r" "A \<noteq> {}"
  shows "r \<le>o r *c |A|"
proof -
  from assms(2) have "Card_order |A|" "\<not> ( |A| =o czero)" by (auto simp add: card_of_empty2)+
  with assms(1) show ?thesis using ordLeq_cprod1 by blast
qed

lemma ordLeq_cprod2[simp]:
  assumes "Cnotzero p1" "Card_order p2"
  shows "p2 \<le>o p1 *c p2"
proof -
  from assms(1) have *: "|Field p1| =o p1" by simp
  { assume "|Field p1| = czero"
    with * have "czero =o p1" using czero_ordIso ordIso_transitive by fastforce
    hence "p1 =o czero" using ordIso_symmetric by blast
    with assms(1) have "False" by blast
  }
  hence "|Field p1| \<noteq> czero" by blast
  hence "Field p1 \<noteq> {}" by (simp add: card_of_czero_iff_empty)
  with assms(2) show ?thesis unfolding cprod_def by (simp add: Card_order_Times2 del: SIGMA_CSUM)
qed

lemma ordLeq_cprod2'[simp]:
  assumes "Card_order r" "A \<noteq> {}"
  shows "r \<le>o |A| *c r"
proof -
  from assms(2) have "Card_order |A|" "\<not> ( |A| =o czero)" by (auto simp add: card_of_empty2)+
  with assms(1) show ?thesis using ordLeq_cprod2 by blast
qed

lemma cinfinite_cprod[simp]:
  "\<lbrakk>cinfinite r1; cinfinite r2\<rbrakk> \<Longrightarrow> cinfinite (r1 *c r2)"
unfolding cinfinite_def cprod_def by (simp del: SIGMA_CSUM)

lemma Cinfinite_cprod[simp]:
  "\<lbrakk>Cinfinite r1; Cinfinite r2\<rbrakk> \<Longrightarrow> Cinfinite (r1 *c r2)"
unfolding cinfinite_def cprod_def by (simp del: SIGMA_CSUM)

lemma cinfinite_cprod1:
  assumes "Cinfinite r1" "Cnotzero r2"
  shows "cinfinite (r1 *c r2)"
proof -
  from assms have "r1 \<le>o r1 *c r2" by (auto intro: ordLeq_cprod1)
  with assms(1) show ?thesis unfolding cinfinite_def cprod_def
    using card_of_mono2 card_of_ordLeq_infinite by auto
qed

lemma Cinfinite_cprod1:
  "\<lbrakk>Cinfinite r1; Cnotzero r2\<rbrakk> \<Longrightarrow> Cinfinite (r1 *c r2)"
by (blast intro: cinfinite_cprod1 Card_order_cprod)

lemma cinfinite_cprod2:
  assumes "Cnotzero r1" "Cinfinite r2"
  shows "cinfinite (r1 *c r2)"
proof -
  from assms have "r2 \<le>o r1 *c r2" by (auto intro: ordLeq_cprod2)
  with assms(2) show ?thesis unfolding cinfinite_def cprod_def
    using card_of_mono2 card_of_ordLeq_infinite by auto
qed

lemma Cinfinite_cprod2:
  "\<lbrakk>Cnotzero r1; Cinfinite r2\<rbrakk> \<Longrightarrow> Cinfinite (r1 *c r2)"
by (blast intro: cinfinite_cprod2 Card_order_cprod)

lemma Cnotzero_cprod:
  assumes r1: "Cnotzero (r1 :: 'a rel)" and r2: "Cnotzero (r2 :: 'b rel)"
  shows "Cnotzero (r1 *c r2)"
proof (intro conjI, rule ccontr)
  assume "\<not> (r1 *c r2, czero :: ('a \<times> 'b) rel) \<notin> ordIso"
  hence "r1 *c r2 =o (czero :: ('a \<times> 'b) rel)" by blast
  hence "Field r1 <*> Field r2 = ({} :: ('a \<times> 'b) set)" unfolding cprod_def
    using iffD1[OF card_of_ordIso_czero_iff_empty] by blast
  hence "Field r1 = {} \<or> Field r2 = {}" by auto
  thus False
    using Cnotzero_imp_not_empty[OF r1] Cnotzero_imp_not_empty[OF r2] by simp
qed simp

lemma cprod_com:
  "p1 *c p2 =o p2 *c p1"
unfolding cprod_def by (simp only: card_of_Times_commute)

lemma cprod_assoc:
  "(p1 *c p2) *c p3 =o p1 *c p2 *c p3"
unfolding cprod_def by (simp only: Field_card_of card_of_Times_assoc)

lemma Prod_cprod[simp]:
  "|A <*> B| =o |A| *c |B|"
unfolding cprod_def by(simp only: Field_card_of card_of_refl)

lemma czero_cprod[simp]:
  "czero *c r =o czero"
unfolding cprod_def czero_def by (simp del: empty_czero add: card_of_empty_ordIso)

lemma cprod_czero[simp]:
  "r *c czero =o czero"
unfolding cprod_def czero_def by (simp del: empty_czero add: card_of_empty_ordIso)

lemma cone_cprod[simp]:
  assumes "Card_order r"
  shows "cone *c r =o r"
proof -
  have "|{()} <*> Field r| =o r" by (simp only: Card_order_Times_singl2 assms ordIso_symmetric)
  thus ?thesis unfolding cprod_def cone_def by (simp only: Field_card_of)
qed

lemma cprod_cone[simp]:
  assumes "Card_order r"
  shows "r *c cone =o r"
proof -
  have "|Field r <*> {()}| =o r" by (simp only: Card_order_Times_singl1 assms ordIso_symmetric)
  thus ?thesis unfolding cprod_def cone_def by (simp only: Field_card_of)
qed

lemma card_of_Csum_Times:
  "\<forall>i \<in> I. |A i| \<le>o |B| \<Longrightarrow> (CSUM i : |I|. |A i| ) \<le>o |I| *c |B|"
unfolding Csum_def cprod_def using card_of_Sigma_Times by simp

lemma card_of_Csum_Times':
  assumes "Card_order r" "\<forall>i \<in> I. |A i| \<le>o r"
  shows "(CSUM i : |I|. |A i| ) \<le>o |I| *c r"
proof -
  from assms(1) have *: "r =o |Field r|" by simp
  with assms(2) have "\<forall>i \<in> I. |A i| \<le>o |Field r|" by (blast intro: ordLeq_ordIso_trans)
  hence "(CSUM i : |I|. |A i| ) \<le>o |I| *c |Field r|" by (simp only: card_of_Csum_Times)
  also from * have "|I| *c |Field r| \<le>o |I| *c r" by simp
  finally show ?thesis .
qed

lemma cprod_csum_distrib1:
  "r1 *c r2 +c r1 *c r3 =o r1 *c (r2 +c r3)"
unfolding csum_def cprod_def
by (simp add: card_of_Times_Plus_distrib ordIso_symmetric del: SIGMA_CSUM)

lemma cprod_csum_distrib2:
  "r1 *c r3 +c r2 *c r3 =o (r1 +c r2) *c r3"
by (blast intro: cprod_com cprod_csum_distrib1 csum_cong ordIso_transitive)

lemma csum_cprod_ordLeq:
  assumes "Cnotzero r1" "Card_order r2"
  shows "r1 +c r2 \<le>o r1 *c (cone +c r2)"
proof -
  from assms(1) have "r1 \<le>o r1 *c cone"
    using Card_order_cone cone_Cnotzero by (blast intro: ordLeq_cprod1)
  moreover from assms have "r2 \<le>o r1 *c r2" by (blast intro: ordLeq_cprod2)
  ultimately have "r1 +c r2 \<le>o r1 *c cone +c r1 *c r2" by (rule csum_mono)
  also have " r1 *c cone +c r1 *c r2 =o r1 *c (cone +c r2)" by (rule cprod_csum_distrib1)
  finally show ?thesis .
qed

lemma csum_absorb2':
  assumes card: "Card_order r2"
  and r12: "r1 \<le>o r2" and cr12: "cinfinite r1 \<or> cinfinite r2"
  shows "r1 +c r2 =o r2"
proof -
  have "infinite (Field r2)"
    using assms card_of_mono2 card_of_ordLeq_infinite
    unfolding csum_def cinfinite_def by auto
  hence "r1 +c r2 =o |Field r2|"
    using card_of_Plus_infinite2 card_of_mono2 r12 unfolding csum_def by auto
  thus ?thesis using card card_of_Field_ordIso ordIso_transitive by auto
qed

lemma csum_absorb2:
  "\<lbrakk>Cinfinite r2; r1 \<le>o r2\<rbrakk> \<Longrightarrow> r1 +c r2 =o r2"
by (rule csum_absorb2') auto

lemma csum_absorb1':
  assumes card: "Card_order r2"
  and r12: "r1 \<le>o r2" and cr12: "cinfinite r1 \<or> cinfinite r2"
  shows "r2 +c r1 =o r2"
by (rule ordIso_transitive, rule csum_com, rule csum_absorb2', (simp only: assms)+)

lemma csum_absorb1:
  "\<lbrakk>Cinfinite r2; r1 \<le>o r2\<rbrakk> \<Longrightarrow> r2 +c r1 =o r2"
by (rule csum_absorb1') auto

lemma cprod_infinite1:
  "\<lbrakk>Cinfinite r; A \<noteq> {}; |A| \<le>o r\<rbrakk> \<Longrightarrow> r *c |A| =o r"
unfolding cinfinite_def cprod_def
by (rule Card_order_Times_infinite[THEN conjunct1]) auto

lemma cprod_infinite1':
  "\<lbrakk>Cinfinite r; Cnotzero p; p \<le>o r\<rbrakk> \<Longrightarrow> r *c p =o r"
unfolding cinfinite_def cprod_def
by (rule Card_order_Times_infinite[THEN conjunct1]) (blast intro: czeroI)+

lemma cprod_infinite1_natLeq:
  "Cinfinite r \<Longrightarrow> r *c natLeq =o r"
unfolding cprod_def
by (rule Card_order_Times_infinite[THEN conjunct1])
   (auto simp add: natLeq_ordLeq_cinfinite cinfinite_def Field_natLeq)

lemma cprod_infinite2:
  "\<lbrakk>Cinfinite r; A \<noteq> {}; |A| \<le>o r\<rbrakk> \<Longrightarrow> |A| *c r =o r"
unfolding cinfinite_def cprod_def
by (rule Card_order_Times_infinite[THEN conjunct2]) auto

lemma cprod_infinite2':
  "\<lbrakk>Cinfinite r; Cnotzero p; p \<le>o r\<rbrakk> \<Longrightarrow> p *c r =o r"
unfolding cinfinite_def cprod_def
by (rule Card_order_Times_infinite[THEN conjunct2]) (blast intro: czeroI)+

lemma cprod_infinite2_natLeq:
  "Cinfinite r \<Longrightarrow> natLeq *c r =o r"
unfolding cprod_def
by (rule Card_order_Times_infinite[THEN conjunct2])
   (auto simp add: natLeq_ordLeq_cinfinite cinfinite_def Field_natLeq)


subsection {* Exponentiation *}

definition cexp (infixr "^c" 80) where
  "r1 ^c r2 \<equiv> |Func (Field r2) (Field r1)|"

definition ccexp (infixr "^^c" 80) where
  "r1 ^^c r2 \<equiv> |Pfunc (Field r2) (Field r1)|"

lemma cexp_ordLeq_ccexp:
  "r1 ^c r2 \<le>o r1 ^^c r2"
unfolding cexp_def ccexp_def by (rule card_of_mono1) (rule Func_Pfunc)

lemma card_order_ccexp:
  assumes "card_order r1" "card_order r2"
  shows "card_order (r1 ^^c r2)"
proof -
  have "Field r1 = UNIV" "Field r2 = UNIV" using assms card_order_on_Card_order by auto
  thus ?thesis unfolding ccexp_def Pfunc_def by (auto split: option.split)
qed

lemma Card_order_cexp: "Card_order (r1 ^c r2)"
unfolding cexp_def by simp

lemma Card_order_ccexp: "Card_order (r1 ^^c r2)"
unfolding ccexp_def by simp

lemma cexp_mono':
  assumes 1: "p1 \<le>o r1" and 2: "p2 \<le>o r2"
  and n1: "Field p1 \<noteq> {} \<or> cone \<le>o r1 ^c r2"
  and n2: "Field p2 = {} \<Longrightarrow> Field r2 = {}"
  shows "p1 ^c p2 \<le>o r1 ^c r2"
proof(cases "Field p1 = {}")
  case True
  hence "|Field (p1 ^c p2)| \<le>o cone"
    unfolding cone_def cexp_def Field_card_of by (cases "Field p2 = {}") auto
  hence "p1 ^c p2 \<le>o cone" by (simp add: cexp_def)
  thus ?thesis using True n1 ordLeq_transitive by auto
next
  case False
  have 1: "|Field p1| \<le>o |Field r1|" and 2: "|Field p2| \<le>o |Field r2|"
    using 1 2 by auto
  obtain f1 where f1: "f1 ` Field r1 = Field p1"
    using 1 unfolding card_of_ordLeq2[OF False, symmetric] by auto
  obtain f2 where f2: "inj_on f2 (Field p2)" "f2 ` Field p2 \<subseteq> Field r2"
    using 2 unfolding card_of_ordLeq[symmetric] by blast
  have 0: "Func_map (Field p2) f1 f2 ` (Field (r1 ^c r2)) = Field (p1 ^c p2)"
    unfolding cexp_def Field_card_of using Func_map_surj[OF f1 f2 n2, symmetric] .
  have 00: "Field (p1 ^c p2) \<noteq> {}" unfolding cexp_def Field_card_of Func_is_emp
    using False by simp
  show ?thesis
    using 0 card_of_ordLeq2[OF 00] unfolding cexp_def Field_card_of by blast
qed

lemma cexp_mono[simp]:
  assumes 1: "p1 \<le>o r1" and 2: "p2 \<le>o r2"
  and n1: "Cnotzero p1 \<or> cone \<le>o r1 ^c r2"
  and n2: "p2 =o czero \<Longrightarrow> r2 =o czero" and card: "Card_order p2"
  shows "p1 ^c p2 \<le>o r1 ^c r2"
proof (rule cexp_mono'[OF 1 2])
  show "Field p1 \<noteq> {} \<or> cone \<le>o r1 ^c r2"
  proof (cases "Cnotzero p1")
    case True show ?thesis using Cnotzero_imp_not_empty[OF True] by (rule disjI1)
  next
    case False with n1 show ?thesis by blast
  qed
qed (rule czeroI[OF card, THEN n2, THEN czeroE])

lemma cexp_mono1':
  assumes 1: "p1 \<le>o r1"
  and n1: "Field p1 \<noteq> {} \<or> cone \<le>o r1 ^c q" and q: "Card_order q"
  shows "p1 ^c q \<le>o r1 ^c q"
using ordLeq_refl[OF q] by (rule cexp_mono'[OF 1 _ n1]) auto

lemma cexp_mono1[simp]:
  assumes 1: "p1 \<le>o r1"
  and n1: "Cnotzero p1 \<or> cone \<le>o r1 ^c q" and q: "Card_order q"
  shows "p1 ^c q \<le>o r1 ^c q"
using ordLeq_refl[OF q] by (rule cexp_mono[OF 1 _ n1]) (auto simp: q)

lemma cexp_mono1_Cnotzero[simp]:
  assumes 1: "p1 \<le>o r1"
  and n1: "Cnotzero p1" and q: "Card_order q"
  shows "p1 ^c q \<le>o r1 ^c q"
using assms by simp

lemma cexp_mono1_cone_ordLeq[simp]:
  assumes 1: "p1 \<le>o r1"
  and n1: "cone \<le>o r1 ^c q" and q: "Card_order q"
  shows "p1 ^c q \<le>o r1 ^c q"
using assms by simp

lemma cexp_mono2':
  assumes 2: "p2 \<le>o r2" and q: "Card_order q"
  and n1: "Field q \<noteq> {} \<or> cone \<le>o q ^c r2"
  and n2: "Field p2 = {} \<Longrightarrow> Field r2 = {}"
  shows "q ^c p2 \<le>o q ^c r2"
using ordLeq_refl[OF q] by (rule cexp_mono'[OF _ 2 n1 n2]) auto

lemma cexp_mono2[simp]:
  assumes 2: "p2 \<le>o r2" and q: "Card_order q"
  and n1: "Cnotzero q \<or> cone \<le>o q ^c r2"
  and n2: "p2 =o czero \<Longrightarrow> r2 =o czero" and card: "Card_order p2"
  shows "q ^c p2 \<le>o q ^c r2"
using ordLeq_refl[OF q] by (rule cexp_mono[OF _ 2 n1 n2 card]) auto

lemma cexp_mono2_Cnotzero[simp]:
  assumes 2: "p2 \<le>o r2"
  and n1: "Cnotzero q" and n2: "Cnotzero p2"
  shows "q ^c p2 \<le>o q ^c r2"
proof (rule cexp_mono)
  assume *: "p2 =o czero"
  have False using n2 czeroI czeroE[OF *] by blast
  thus "r2 =o czero" by blast
qed (auto simp add: assms ordLeq_refl)

lemma cexp_cong':
  assumes 1: "p1 =o r1" and 2: "p2 =o r2"
  and p1: "Field p1 \<noteq> {} \<or> cone \<le>o r1 ^c r2"
  and r1: "Field r1 \<noteq> {} \<or> cone \<le>o p1 ^c p2"
  shows "p1 ^c p2 =o r1 ^c r2"
proof -
  obtain f where "bij_betw f (Field p2) (Field r2)"
    using 2 card_of_ordIso[of "Field p2" "Field r2"] card_of_cong by auto
  hence 0: "Field p2 = {} \<longleftrightarrow> Field r2 = {}" unfolding bij_betw_def by auto
  show ?thesis using 0 1 2 cexp_mono'[OF _ _ p1] cexp_mono'[OF _ _ r1]
    unfolding ordIso_iff_ordLeq by auto
qed

lemma cexp_cong[simp]:
  assumes 1: "p1 =o r1" and 2: "p2 =o r2"
  and p1: "Cnotzero p1 \<or> cone \<le>o r1 ^c r2" and Cr: "Card_order r2"
  and r1: "Cnotzero r1 \<or> cone \<le>o p1 ^c p2" and Cp: "Card_order p2"
  shows "p1 ^c p2 =o r1 ^c r2"
proof -
  obtain f where "bij_betw f (Field p2) (Field r2)"
    using 2 card_of_ordIso[of "Field p2" "Field r2"] card_of_cong by auto
  hence 0: "Field p2 = {} \<longleftrightarrow> Field r2 = {}" unfolding bij_betw_def by auto
  have r: "p2 =o czero \<Longrightarrow> r2 =o czero"
    and p: "r2 =o czero \<Longrightarrow> p2 =o czero"
     using 0 Cr Cp czeroE czeroI by auto
  show ?thesis using 0 1 2 unfolding ordIso_iff_ordLeq
    using r p cexp_mono[OF _ _ p1 _ Cp] cexp_mono[OF _ _ r1 _ Cr]
    by blast
qed

lemma cexp_cong1':
  assumes 1: "p1 =o r1" and q: "Card_order q"
  and p1: "Field p1 \<noteq> {} \<or> cone \<le>o r1 ^c q"
  and r1: "Field r1 \<noteq> {} \<or> cone \<le>o p1 ^c q"
  shows "p1 ^c q =o r1 ^c q"
by (rule cexp_cong'[OF 1 _ p1 r1]) (rule ordIso_refl[OF q])

lemma cexp_cong1[simp]:
  assumes 1: "p1 =o r1" and q: "Card_order q"
  and p1: "Cnotzero p1 \<or> cone \<le>o r1 ^c q"
  and r1: "Cnotzero r1 \<or> cone \<le>o p1 ^c q"
  shows "p1 ^c q =o r1 ^c q"
by (rule cexp_cong[OF 1 _ p1 q r1 q]) (rule ordIso_refl[OF q])

lemma cexp_cong1_Cnotzero[simp]:
  assumes 1: "p1 =o r1" and q: "Card_order q"
  and p1: "Cnotzero p1" and r1: "Cnotzero r1"
  shows "p1 ^c q =o r1 ^c q"
by (rule cexp_cong1, auto simp add: assms)

lemma cexp_cong1_cone_ordLeq[simp]:
  assumes 1: "p1 =o r1" and q: "Card_order q"
  and p1: "cone \<le>o r1 ^c q" and r1: "cone \<le>o p1 ^c q"
  shows "p1 ^c q =o r1 ^c q"
by (rule cexp_cong1, auto simp add: assms)

lemma cexp_cong2':
  assumes 2: "p2 =o r2" and q: "Card_order q"
  shows "Field q \<noteq> {} \<or> (cone \<le>o q ^c p2 \<and> cone \<le>o q ^c r2) \<Longrightarrow>
    q ^c p2 =o q ^c r2"
by (rule cexp_cong'[OF _ 2]) (auto simp only: ordIso_refl q)

lemma cexp_cong2[simp]:
  assumes 2: "p2 =o r2" and q: "Card_order q"
  and p: "Card_order p2" and r: "Card_order r2"
  shows "Cnotzero q \<or> (cone \<le>o q ^c p2 \<and> cone \<le>o q ^c r2) \<Longrightarrow>
    q ^c p2 =o q ^c r2"
by (rule cexp_cong[OF _ 2]) (auto simp only: ordIso_refl q p r)

lemma cexp_cong2_Cnotzero[simp]:
  assumes 2: "p2 =o r2" and q: "Cnotzero q"
  and p: "Card_order p2"
  shows "q ^c p2 =o q ^c r2"
by (rule cexp_cong[OF _ 2]) (auto simp only: ordIso_refl Card_order_ordIso2[OF p 2] q p)

lemma cexp_czero:
  "r ^c czero =o cone"
unfolding cexp_def czero_def Field_card_of Func_emp_empty by (rule single_cone)

lemma czero_cexp:
  "Cnotzero r \<Longrightarrow> czero ^c r =o czero"
unfolding cexp_def czero_def Field_card_of
using Func_emp2[of "Field r"] card_of_empty_ordIso card_of_unique[of "Field r" r]
by fastforce

lemma cexp_cone:
  assumes "Card_order r"
  shows "r ^c cone =o r"
proof -
  have "r ^c cone =o |Field r|"
    unfolding cexp_def cone_def Field_card_of Func_emp_empty
      card_of_ordIso[symmetric] bij_betw_def Func_def inj_on_def image_def
    by (rule exI[of _ "\<lambda>f. case f () of Some a \<Rightarrow> a"]) auto
  also have "|Field r| =o r" by (rule card_of_Field_ordIso[OF assms])
  finally show ?thesis .
qed

lemma cexp_cprod:
  assumes r1: "Cnotzero r1"
  shows "(r1 ^c r2) ^c r3 =o r1 ^c (r2 *c r3)" (is "?L =o ?R")
proof -
  have "?L =o r1 ^c (r3 *c r2)"
    unfolding cprod_def cexp_def Field_card_of
    using card_of_Func_Times by(rule ordIso_symmetric)
  also have "r1 ^c (r3 *c r2) =o ?R"
    apply(rule cexp_cong2) using cprod_com r1 by auto
  finally show ?thesis .
qed

lemma cexp_cprod_ordLeq:
  assumes r1: "Cnotzero r1" and r2: "Cinfinite r2"
  and r3: "Cnotzero r3" "r3 \<le>o r2"
  shows "(r1 ^c r2) ^c r3 =o r1 ^c r2" (is "?L =o ?R")
proof-
  have "?L =o r1 ^c (r2 *c r3)" using cexp_cprod[OF r1] .
  also have "r1 ^c (r2 *c r3) =o ?R"
  apply(rule cexp_cong2)
  apply(rule cprod_infinite1'[OF r2 r3]) using r1 r2 by fastforce+
  finally show ?thesis .
qed

lemma cexp_cprod_natLeq:
  assumes r1: "Cnotzero r1"
  and r2: "Cinfinite r2"
  shows "(r1 ^c r2) ^c natLeq =o r1 ^c r2" (is "?L =o ?R")
proof-
  have "?L =o r1 ^c (r2 *c natLeq)" using cexp_cprod[OF r1] .
  also have "r1 ^c (r2 *c natLeq) =o ?R"
  apply(rule cexp_cong2)
  apply(rule cprod_infinite1_natLeq[OF r2]) using r1 r2 by fastforce+
  finally show ?thesis .
qed

lemma Cnotzero_UNIV[simp]: "Cnotzero |UNIV|"
by (simp, rule notI, drule czeroE, auto)

lemma card_of_Pfunc_Func:
  assumes A: "infinite A" and B: "{b1,b2} \<subseteq> B" "b1 \<noteq> b2"
  shows "|Pfunc A B| =o |Func A B|"
unfolding ordIso_iff_ordLeq proof
  show "|Func A B| \<le>o |Pfunc A B|" using Func_Pfunc by (rule card_of_mono1)
next
  have B': "B \<noteq> {}" using B by auto
  have B'': "|UNIV::bool set| \<le>o |B|"
    using ordIso_ordLeq_trans[OF card_of_bool[OF B(2)] card_of_mono1[OF B(1)]] .
  have "|Pfunc A B| \<le>o |Pow A \<times> Func A B|" by (rule card_of_Pfunc_Pow_Func[OF B'])
  also have "|Pow A \<times> Func A B| =o |Func A (UNIV :: bool set) \<times> Func A B|"
    by (rule card_of_Times_cong1[OF card_of_Pow_Func])
  also have "|Func A (UNIV :: bool set) \<times> Func A B| =o |Func A B|"
  proof (rule card_of_Times_infinite[THEN conjunct2, OF infinite_Func[OF A B]])
    show "Func A (UNIV :: bool set) \<noteq> {}" using A unfolding Func_is_emp by simp
  next
    show "|Func A (UNIV :: bool set)| \<le>o |Func A B|"
      using cexp_mono1[OF B'', of "|A|", OF _ card_of_Card_order]
      unfolding cexp_def Field_card_of by auto
  qed
  finally show "|Pfunc A B| \<le>o |Func A B|" .
qed

lemma Pow_cexp_ctwo:
  "|Pow A| =o ctwo ^c |A|"
unfolding ctwo_def cexp_def Field_card_of by (rule card_of_Pow_Func)

lemma ordLess_ctwo_cexp:
  assumes "Card_order r"
  shows "r <o ctwo ^c r"
proof -
  have "r <o |Pow (Field r)|" using assms by simp
  also have "|Pow (Field r)| =o ctwo ^c r"
    unfolding ctwo_def cexp_def Field_card_of by (rule card_of_Pow_Func)
  finally show ?thesis .
qed

lemma ordLeq_cexp1:
  assumes "Cnotzero r" "Card_order q"
  shows "q \<le>o q ^c r"
proof (cases "q =o (czero :: 'a rel)")
  case True thus ?thesis by (simp add: ordIso_ordLeq_trans Card_order_cexp)
next
  case False
  thus ?thesis
    apply -
    apply (rule ordIso_ordLeq_trans)
    apply (rule ordIso_symmetric)
    apply (rule cexp_cone)
    apply (rule assms(2))
    apply (rule cexp_mono2)
    apply (rule cone_ordLeq_Cnotzero)
    apply (rule assms(1))
    apply (rule assms(2))
    apply (rule disjI1)
    apply (rule conjI)
    apply (rule notI)
    apply (erule notE)
    apply (rule ordIso_transitive)
    apply assumption
    apply (rule czero_ordIso)
    apply (rule assms(2))
    apply (rule notE)
    apply (rule cone_not_czero)
    apply assumption
    apply (rule Card_order_cone)
  done
qed

lemma ordLeq_cexp2:
  assumes "ctwo \<le>o q" "Card_order r"
  shows "r \<le>o q ^c r"
proof (cases "r =o (czero :: 'a rel)")
  case True thus ?thesis by (simp add: ordIso_ordLeq_trans Card_order_cexp)
next
  case False thus ?thesis
    apply -
    apply (rule ordLess_imp_ordLeq)
    apply (rule ordLess_ordLeq_trans)
    apply (rule ordLess_ctwo_cexp)
    apply (rule assms(2))
    apply (rule cexp_mono1)
    apply (rule assms(1))
    apply (rule disjI1)
    apply (rule ctwo_Cnotzero)
    apply (rule assms(2))
  done
qed

lemma Cnotzero_cexp:
  assumes "Cnotzero q" "Card_order r"
  shows "Cnotzero (q ^c r)"
proof (cases "r =o czero")
  case False thus ?thesis
    apply -
    apply (rule Cnotzero_mono)
    apply (rule assms(1))
    apply (rule Card_order_cexp)
    apply (rule ordLeq_cexp1)
    apply (rule conjI)
    apply (rule notI)
    apply (erule notE)
    apply (rule ordIso_transitive)
    apply assumption
    apply (rule czero_ordIso)
    apply (rule assms(2))
    apply (rule conjunct2)
    apply (rule assms(1))
  done
next
  case True thus ?thesis
    apply -
    apply (rule Cnotzero_mono)
    apply (rule cone_Cnotzero)
    apply (rule Card_order_cexp)
    apply (rule ordIso_imp_ordLeq)
    apply (rule ordIso_symmetric)
    apply (rule ordIso_transitive)
    apply (rule cexp_cong2)
    apply assumption
    apply (rule conjunct2)
    apply (rule assms(1))
    apply (rule assms(2))
    apply (simp only: card_of_Card_order czero_def)
    apply (rule disjI1)
    apply (rule assms(1))
    apply (rule cexp_czero)
  done
qed

lemma cinfinite_ctwo_cexp:
  "cinfinite r \<Longrightarrow> cinfinite (ctwo ^c r)"
unfolding ctwo_def cexp_def cinfinite_def Field_card_of
by (rule infinite_Func) auto

lemma Cinfinite_ctwo_cexp:
  "Cinfinite r \<Longrightarrow> Cinfinite (ctwo ^c r)"
unfolding ctwo_def cexp_def cinfinite_def Field_card_of
by (rule conjI, rule infinite_Func) auto

lemma cinfinite_cexp:
  assumes q: "ctwo \<le>o q" and r: "Cinfinite r"
  shows "cinfinite (q ^c r)"
proof -
  have "ctwo ^c r \<le>o q ^c r" by (rule cexp_mono1[OF q]) (auto simp add: r ctwo_def)
  thus ?thesis using Cinfinite_ctwo_cexp[OF r]
  unfolding cinfinite_def using card_of_ordLeq_infinite card_of_mono2 by metis
qed

lemma cinfinite_ccexp:
  "\<lbrakk>ctwo \<le>o q; Cinfinite r\<rbrakk> \<Longrightarrow> cinfinite (q ^^c r)"
using cinfinite_mono[OF cexp_ordLeq_ccexp cinfinite_cexp] by auto

lemma Cinfinite_cexp:
  "\<lbrakk>ctwo \<le>o q; Cinfinite r\<rbrakk> \<Longrightarrow> Cinfinite (q ^c r)"
by (simp add: cinfinite_cexp Card_order_cexp)

lemma Cinfinite_ccexp:
  "\<lbrakk>ctwo \<le>o q; Cinfinite r\<rbrakk> \<Longrightarrow> Cinfinite (q ^^c r)"
using Cinfinite_mono[OF cexp_ordLeq_ccexp Cinfinite_cexp]
by (auto simp add: Card_order_ccexp)

lemma ctwo_ordLeq_natLeq:
  "ctwo \<le>o natLeq"
proof -
  have "ctwo \<le>o |Field natLeq|"
    unfolding Field_natLeq ctwo_def inj_on_def card_of_ordLeq[symmetric]
    by (rule exI[of _ "\<lambda> c. if c then 0 else Suc 0"]) auto
  thus ?thesis using card_of_Field_natLeq ordLeq_ordIso_trans by blast
qed

lemma ctwo_ordLess_natLeq:
  "ctwo <o natLeq"
unfolding ctwo_def using finite_iff_ordLess_natLeq finite_UNIV by fastforce

lemma ctwo_ordLess_Cinfinite:
  assumes r: "Cinfinite r"
  shows "ctwo <o r"
proof -
  have "natLeq \<le>o r" using r by (rule natLeq_ordLeq_cinfinite)
  thus ?thesis using ctwo_ordLess_natLeq ordLess_ordLeq_trans by blast
qed

lemma ctwo_ordLeq_Cinfinite:
  assumes r: "Cinfinite r"
  shows "ctwo \<le>o r"
by (rule ordLess_imp_ordLeq[OF ctwo_ordLess_Cinfinite[OF r]])

lemma Cinfinite_ordLess_cexp:
  assumes r: "Cinfinite r"
  shows "r <o r ^c r"
proof -
  have "r <o ctwo ^c r" using r by (simp only: ordLess_ctwo_cexp)
  also have "ctwo ^c r \<le>o r ^c r"
    by (rule cexp_mono1[OF ctwo_ordLeq_Cinfinite]) (auto simp: r ctwo_not_czero)
  finally show ?thesis .
qed

lemma infinite_ordLeq_cexp:
  assumes r: "Cinfinite r"
  shows "r \<le>o r ^c r"
by (rule ordLess_imp_ordLeq[OF Cinfinite_ordLess_cexp[OF r]])

lemma cone_ordLeq_iff_Field:
  assumes "cone \<le>o r"
  shows "Field r \<noteq> {}"
proof (rule ccontr)
  assume "\<not> Field r \<noteq> {}"
  hence "Field r = {}" by simp
  thus False using card_of_empty3
    card_of_mono2[OF assms] Cnotzero_imp_not_empty[OF cone_Cnotzero] by auto
qed

lemma cone_ordLeq_cexp[simp]:
  assumes "cone \<le>o r1"
  shows "cone \<le>o r1 ^c r2"
proof -
  have "Field r1 \<noteq> {}" using assms cone_ordLeq_iff_Field by auto
  thus ?thesis unfolding cone_def cexp_def by auto
qed

lemma Card_order_czero[simp]: "Card_order czero"
by (simp only: card_of_Card_order czero_def)

lemma cexp_mono2'':
  assumes 2: "p2 \<le>o r2"
  and n1: "Cnotzero q"
  and n2: "Card_order p2"
  shows "q ^c p2 \<le>o q ^c r2"
proof (cases "p2 =o (czero :: 'a rel)")
  case True
  hence "q ^c p2 =o q ^c (czero :: 'a rel)" using n1 n2 cexp_cong2 Card_order_czero by blast
  also have "q ^c (czero :: 'a rel) =o cone" using cexp_czero by blast
  also have "cone \<le>o q ^c r2" using cone_ordLeq_cexp cone_ordLeq_Cnotzero n1 by blast
  finally show ?thesis .
next
  case False thus ?thesis using assms cexp_mono2' czeroI by metis
qed


subsection {* Infinite bounds *}

lemma cone_ordLeq_cinfinite: "Cinfinite r \<Longrightarrow> cone \<le>o r"
unfolding cinfinite_def cone_def
by (auto intro: Card_order_singl_ordLeq)

lemma Un_Cinfinite_bound:
  "\<lbrakk>|A| \<le>o r; |B| \<le>o r; Cinfinite r\<rbrakk> \<Longrightarrow> |A \<union> B| \<le>o r"
by (auto simp add: cinfinite_def)

lemma UNION_Cinfinite_bound:
  "\<lbrakk>|I| \<le>o r; \<forall>i \<in> I. |A i| \<le>o r; Cinfinite r\<rbrakk> \<Longrightarrow> |\<Union>i \<in> I. A i| \<le>o r"
by (auto simp add: card_of_UNION_ordLeq_infinite_Field cinfinite_def)

lemma csum_cinfinite_bound:
  assumes "p \<le>o r" "q \<le>o r" "Card_order p" "Card_order q" "Cinfinite r"
  shows "p +c q \<le>o r"
proof -
  from assms(1-4) have "|Field p| \<le>o r" "|Field q| \<le>o r"
    unfolding card_order_on_def
    using card_of_least ordLeq_transitive by blast+
  with assms show ?thesis unfolding cinfinite_def csum_def
    by (blast intro: card_of_Plus_ordLeq_infinite_Field)
qed

lemma csum_cexp: "\<lbrakk>Cinfinite r1; Cinfinite r2; Card_order q; ctwo \<le>o q\<rbrakk> \<Longrightarrow>
  q ^c r1 +c q ^c r2 \<le>o q ^c (r1 +c r2)"
apply (rule csum_cinfinite_bound)
apply (rule cexp_mono2)
apply (rule ordLeq_csum1)
apply (erule conjunct2)
apply assumption
apply (rule disjI2)
apply (rule ordLeq_transitive)
apply (rule cone_ordLeq_ctwo)
apply (rule ordLeq_transitive)
apply assumption
apply (rule ordLeq_cexp1)
apply (rule Cinfinite_Cnotzero)
apply (rule Cinfinite_csum)
apply (rule disjI1)
apply assumption
apply assumption
apply (rule notE)
apply (rule cinfinite_not_czero[of r1])
apply (erule conjunct1)
apply assumption
apply (erule conjunct2)
apply (rule cexp_mono2)
apply (rule ordLeq_csum2)
apply (erule conjunct2)
apply assumption
apply (rule disjI2)
apply (rule ordLeq_transitive)
apply (rule cone_ordLeq_ctwo)
apply (rule ordLeq_transitive)
apply assumption
apply (rule ordLeq_cexp1)
apply (rule Cinfinite_Cnotzero)
apply (rule Cinfinite_csum)
apply (rule disjI1)
apply assumption
apply assumption
apply (rule notE)
apply (rule cinfinite_not_czero[of r2])
apply (erule conjunct1)
apply assumption
apply (erule conjunct2)
apply (rule Card_order_cexp)
apply (rule Card_order_cexp)
apply (rule Cinfinite_cexp)
apply assumption
apply (rule Cinfinite_csum)
apply (rule disjI1)
apply assumption
done

lemma csum_cexp': "\<lbrakk>Cinfinite r; Card_order q; ctwo \<le>o q\<rbrakk> \<Longrightarrow> q +c r \<le>o q ^c r"
by (rule csum_cinfinite_bound) (auto simp add: ordLeq_cexp1 ordLeq_cexp2 Cinfinite_cexp)

lemma cprod_cinfinite_bound:
  assumes "p \<le>o r" "q \<le>o r" "Card_order p" "Card_order q" "Cinfinite r"
  shows "p *c q \<le>o r"
proof -
  from assms(1-4) have "|Field p| \<le>o r" "|Field q| \<le>o r"
    unfolding card_order_on_def
    using card_of_least ordLeq_transitive by blast+
  with assms show ?thesis unfolding cinfinite_def cprod_def
    by (blast intro: card_of_Times_ordLeq_infinite_Field)
qed

lemma cprod_csum_cexp:
  "r1 *c r2 \<le>o (r1 +c r2) ^c ctwo"
unfolding cprod_def csum_def cexp_def ctwo_def Field_card_of
proof -
  let ?f = "\<lambda>(a, b). %x. if x then Some (Inl a) else Some (Inr b)"
  have "inj_on ?f (Field r1 \<times> Field r2)" (is "inj_on _ ?LHS")
    unfolding inj_on_def fun_eq_iff by (auto split: bool.split)
  moreover
  have "?f ` ?LHS \<subseteq> Func (UNIV :: bool set) (Field r1 <+> Field r2)" (is "_ \<subseteq> ?RHS")
    unfolding Func_def by auto
  ultimately show "|?LHS| \<le>o |?RHS|" using card_of_ordLeq by blast
qed


(* cardSuc *)

lemma Cinfinite_cardSuc: "Cinfinite r \<Longrightarrow> Cinfinite (cardSuc r)"
by (rule Cinfinite_mono) auto

lemma Cnotzero_cardSuc: "Card_order r \<Longrightarrow> Cnotzero (cardSuc r)"
using Field_cardSuc_not_empty czeroE by auto

lemma cardSuc_UNION_Cinfinite:
assumes r: "Cinfinite r"
and As: "relChain (cardSuc r) As"
and Bsub: "B \<le> (UN i : Field (cardSuc r). As i)"
and cardB: "|B| <=o r"
shows "EX i : Field (cardSuc r). B \<le> As i"
using cardSuc_UNION assms unfolding cinfinite_def by blast

subsection {* Powerset *}

definition cpow where "cpow r = |Pow (Field r)|"

lemma card_order_cpow[simp]:
  assumes "card_order r"
  shows "card_order (cpow r)"
proof -
  have "Field r = UNIV" using assms card_order_on_Card_order by auto
  thus ?thesis unfolding cpow_def by auto
qed

lemma cpow_greater[simp]: "Card_order r \<Longrightarrow> r <o cpow r"
unfolding cpow_def by auto

lemma cpow_greater_eq[simp]: "Card_order r \<Longrightarrow> r \<le>o cpow r"
by (rule ordLess_imp_ordLeq) (erule cpow_greater)

lemma Card_order_cpow[simp]: "Card_order (cpow r)"
unfolding cpow_def by simp

lemma Cinfinite_cpow: "Cinfinite r \<Longrightarrow> Cinfinite (cpow r)"
by (rule Cinfinite_mono) auto

lemma Cnotzero_cpow: "Card_order r \<Longrightarrow> Cnotzero (cpow r)"
unfolding cpow_def by (auto , drule czeroE, auto)

lemma cardSuc_ordLeq_cpow: "Card_order r \<Longrightarrow> cardSuc r \<le>o cpow r"
by (rule cardSuc_least) auto

lemma cpow_cexp_ctwo:
  "cpow r =o ctwo ^c r"
unfolding cpow_def ctwo_def cexp_def Field_card_of by (rule card_of_Pow_Func)

end
